#pragma once

#include <iostream>
#include <vector>

#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <elcorenn/elcorenn.h>

#include "coco_classes.hpp"
#include "nms.hpp"

struct Detection {
  int class_id;
  float confidence;
  cv::Rect box;
};

void output_proc(cv::Mat& input_image, std::vector<Detection>& preds,
                 const std::vector<std::pair<std::string, cv::Scalar>>& classes,
                 const double score_threshold, const double nms_threshold,
                 const double confidence_threshold) {
  int rows = input_image.rows;
  int dimensions = input_image.cols;

  bool yolov5 = true;
  if (dimensions > rows) {
    yolov5 = false;

    rows = input_image.cols;
    dimensions = input_image.rows;

    input_image = input_image.reshape(1, dimensions);
    cv::transpose(input_image, input_image);
  }

  float* data = reinterpret_cast<float*>(input_image.data);

  std::vector<int> class_ids;
  std::vector<float> confidences;
  std::vector<cv::Rect> boxes;

  for (int i = 0; i < rows; ++i) {
    if (yolov5) {
      float confidence = data[4];

      if (confidence >= confidence_threshold) {
        float* classes_scores = data + 5;
        cv::Mat scores(1, classes.size(), CV_32FC1, classes_scores);

        cv::Point class_id;
        double maxClassScore;

        minMaxLoc(scores, 0, &maxClassScore, 0, &class_id);

        if (maxClassScore > score_threshold) {
          confidences.push_back(confidence);
          class_ids.push_back(class_id.x);

          float x = data[0];
          float y = data[1];
          float w = data[2];
          float h = data[3];
          int left = x - 0.5 * w;
          int top = y - 0.5 * h;

          boxes.push_back(cv::Rect(left, top, w, h));
        }
      }
    } else {
      float* classes_scores = data + 4;

      cv::Mat scores(1, classes.size(), CV_32FC1, classes_scores);

      cv::Point class_id;
      double maxClassScore;

      minMaxLoc(scores, 0, &maxClassScore, 0, &class_id);

      if (maxClassScore > score_threshold) {
        confidences.push_back(maxClassScore);
        class_ids.push_back(class_id.x);

        float x = data[0];
        float y = data[1];
        float w = data[2];
        float h = data[3];
        int left = x - 0.5 * w;
        int top = y - 0.5 * h;

        boxes.push_back(cv::Rect(left, top, w, h));
      }
    }

    data += dimensions;
  }

  std::vector<int> nms_result;
  NMSBoxes(boxes, confidences, score_threshold, nms_threshold, nms_result);

  for (int i = 0; i < nms_result.size(); ++i) {
    int idx = nms_result[i];
    Detection result;
    result.class_id = class_ids[idx];
    result.confidence = confidences[idx];
    result.box = boxes[idx];
    preds.push_back(result);
  }
}
